#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Generic Attributes.

These are handy and type save key-value-pairs.

    >>> cast_attrs({'a': '1', 'b': '2'})
    (Attr('a', value='1'), Attr('b', value='2'))
    >>> cast_attrs(("a=1", "b=2"))
    (Attr('a', value='1'), Attr('b', value='2'))
    >>> cast_attrs("a=1; b=2")
    (Attr('a', value='1'), Attr('b', value='2'))
    >>> as_dict(cast_attrs("a=1; b=2"))
    {'a': '1', 'b': '2'}
    >>> as_tuple(cast_attrs("a=1; b=2"))
    (('a', '1'), ('b', '2'))
"""

from typing import Annotated, TypeAlias, Union

import ucdp as u
from pydantic import (
    BeforeValidator,
)


class Attr(u.IdentLightObject):
    """
    Attribute.

    Immutable Key-Value Pair.
    """

    value: str | None = None

    _posargs: u.ClassVar[u.PosArgs] = ("name",)

    def __init__(self, name: str, value: str | None = None) -> None:
        super().__init__(name=name, value=value)

    def __str__(self) -> str:
        if self.value is None:
            return self.name
        return f"{self.name}={self.value}"

    @staticmethod
    def cast(attr: Union["Attr", str]) -> "Attr":
        """
        Cast To Attribute.

            >>> Attr.cast(Attr('one'))
            Attr('one')
            >>> Attr.cast("one")
            Attr('one')
            >>> Attr.cast("one=2")
            Attr('one', value='2')

        """
        if isinstance(attr, Attr):
            return attr
        if "=" in attr:
            name, value = attr.split("=", 1)
            return Attr(name=name, value=value)
        return Attr(name=attr)


Attrs: TypeAlias = tuple[Attr, ...]


def cast_attrs(attrs: Union["Attrs", u.Names, dict, None]) -> "Attrs":
    """
    Cast To Attributes.

        >>> cast_attrs({'a': '1', 'b': '2'})
        (Attr('a', value='1'), Attr('b', value='2'))
        >>> cast_attrs(("a=1", "b=2"))
        (Attr('a', value='1'), Attr('b', value='2'))
        >>> cast_attrs("a=1; b=2")
        (Attr('a', value='1'), Attr('b', value='2'))
        >>> cast_attrs("")
        ()
        >>> cast_attrs(None)
        ()
    """
    if not attrs:
        return ()
    if isinstance(attrs, dict):
        attrs = tuple(Attr(name, value=value) for name, value in attrs.items())
    if not isinstance(attrs, tuple) or not all(isinstance(attr, Attr) for attr in attrs):
        attrs = tuple(Attr.cast(attr) for attr in u.split(attrs))
    # check for duplicates
    if len(set(dict(attrs))) != len(attrs):
        raise ValueError(f"Duplicates in {as_str(attrs)!r}")
    return attrs


def as_str(attrs: Attrs) -> str:
    """
    Return Attributes As String.

        >>> attrs = (Attr('a', value='1'), Attr('b'))
        >>> as_str(attrs)
        'a=1; b'

    """
    return "; ".join(str(attr) for attr in attrs)


format_attrs = as_str


def as_dict(attrs: Attrs) -> dict[str, str | None]:
    """Return Attributes As Dictionary."""
    return {attr.name: attr.value for attr in attrs}


def as_tuple(attrs: Attrs) -> tuple[tuple[str, str | None], ...]:
    """Return Attributes As Tuple."""
    return tuple((attr.name, attr.value) for attr in attrs)


CastableAttrs = Annotated[
    Attrs,
    BeforeValidator(lambda x: cast_attrs(x)),
]



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Generic Lanes.

A lane is anything which is related to any size and optionally additional attributes.
"""

from typing import TypeAlias

import ucdp as u

from .attrs import CastableAttrs


class Lane(u.IdentLightObject):
    """Lane."""

    size: u.Bytesize | None = None
    """Size in Bytes."""

    attrs: CastableAttrs = u.Field(default_factory=dict)
    """Attributes."""


class DefaultLane(Lane):
    """Default Lane."""

    name: str = u.Field(init=False, default="")


Lanes: TypeAlias = tuple[Lane, ...]


def fill_lanes(lanes: Lanes, size: u.Bytesize, default: bool = False) -> Lanes:
    """
    Fill Empty Lane Sizes.

        >>> fill_lanes((Lane(name='a'), Lane(name='b')), u.Bytesize('48kB'))
        (Lane(name='a', size=Bytesize('24 KB')), Lane(name='b', size=Bytesize('24 KB')))
        >>> fill_lanes((Lane(name='a', size='24k'), Lane(name='b', size='24k')), u.Bytesize('48kB'))
        (Lane(name='a', size=Bytesize('24 KB')), Lane(name='b', size=Bytesize('24 KB')))

        >>> fill_lanes((Lane(name='a', size='48k'), Lane(name='b')), u.Bytesize('48kB'))
        Traceback (most recent call last):
            ...
        ValueError: Lanes (Lane(name='a', size=Bytesize('48 KB')), Lane(name='b')) exceed size 48 KB
        >>> fill_lanes((Lane(name='a', size='24k'), Lane(name='b', size='25k')), u.Bytesize('48kB'))
        Traceback (most recent call last):
            ...
        ValueError: Lanes (Lane(name='a', ... exceed size 48 KB

    Empty lanes:

        >>> fill_lanes((), u.Bytesize('48kB'))
        ()
        >>> fill_lanes((), u.Bytesize('48kB'), default=True)
        (DefaultLane(size=Bytesize('48 KB')),)
    """
    if not lanes and default:
        return (DefaultLane(size=size),)
    used = 0
    missing = 0
    for lane in lanes:
        if lane.size is None:
            missing += 1
        else:
            used += lane.size
    if used > size:
        raise ValueError(f"Lanes {lanes} exceed size {size}")
    if not missing:
        return lanes
    lanesize = (size - used) // missing
    if lanesize < 1:
        raise ValueError(f"Lanes {lanes} exceed size {size}")
    return tuple(lane.new(size=lanesize) if lane.size is None else lane for lane in lanes)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Test Attr.
"""

import re

import ucdp as u
from pytest import raises

from ucdp_glbl.attrs import Attr, Attrs, cast_attrs


def test_attr():
    """Attr Without Value."""
    attr = Attr("one")
    assert attr.name == "one"
    assert attr.value is None


def test_attr_value():
    """Attr With Value."""
    attr = Attr("one", value="8k")
    assert attr.name == "one"
    assert attr.value == "8k"


def test_name():
    """Name."""
    with raises(u.ValidationError):
        Attr("1two")
    with raises(u.ValidationError):
        Attr("two1 f")
    with raises(u.ValidationError):
        Attr("two%")


def test_cast():
    """Casting."""
    assert cast_attrs("") == Attrs()
    assert cast_attrs("b") == (Attr("b"),)
    assert cast_attrs("b=1") == (Attr("b", value="1"),)
    assert cast_attrs("b=1;c") == (Attr("b", value="1"), Attr("c"))

    msg = "Duplicates in 'b=1; b=2'"
    with raises(ValueError, match=re.escape(msg)):
        cast_attrs("b=1;b=2")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Test Lanes.
"""

from ucdp_glbl.attrs import Attr, Attrs
from ucdp_glbl.lane import Lane


def test_lane():
    """Lane Basics."""
    lane = Lane(name="one", size="8k")
    assert lane.name == "one"
    assert lane.size == 8 * 1024
    assert lane.attrs == Attrs()

    assert hash(lane)


def test_lane_attrs():
    """Lane Attributes."""
    assert Lane(name="a", size=1, attrs="").attrs == Attrs()
    assert Lane(name="a", size=1, attrs="b").attrs == (Attr("b"),)
    assert Lane(name="a", size=1, attrs="b=1").attrs == (Attr("b", value="1"),)
    assert Lane(name="a", size=1, attrs="b=1;c").attrs == (Attr("b", value="1"), Attr("c"))



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Generic Design For Test Types.
"""

import ucdp as u


class TestModeType(u.AEnumType):
    """
    Test Mode.

    General Chip Test Mode.

        >>> for item in TestModeType().values(): print(item)
        EnumItem(0, 'func', doc=Doc(title='Functional Mode'))
        EnumItem(1, 'test', doc=Doc(title='Test Mode'))
    """

    keytype: u.BitType = u.BitType()

    comment: str = "Test Mode"

    def _build(self) -> None:
        self._add(0, "func", "Functional Mode")
        self._add(1, "test", "Test Mode")


class ScanModeType(u.AEnumType):
    """
    Scan Mode.

    Scan Mode of Digital Logic.

        >>> for item in ScanModeType().values(): print(item)
        EnumItem(0, 'func', doc=Doc(title='Functional Mode'))
        EnumItem(1, 'scan', doc=Doc(title='Scan Mode'))
    """

    keytype: u.BitType = u.BitType()

    comment: str = "Logic Scan-Test Mode"

    def _build(self) -> None:
        self._add(0, "func", "Functional Mode")
        self._add(1, "scan", "Scan Mode")


class ScanShiftType(u.AEnumType):
    """
    Scan Shift.

    Shift Mode During Digital Logic Test.

        >>> for item in ScanShiftType().values(): print(item)
        EnumItem(0, 'capture', doc=Doc(title='Capture Phase during Scan Mode'))
        EnumItem(1, 'shift', doc=Doc(title='Shift Phase during Scan Mode'))
    """

    keytype: u.BitType = u.BitType()

    comment: str = "Scan Shift Phase"

    def _build(self) -> None:
        self._add(0, "capture", "Capture Phase during Scan Mode")
        self._add(1, "shift", "Shift Phase during Scan Mode")


class MbistModeType(u.AEnumType):
    """
    Memory BIST Mode.

        >>> for item in MbistModeType().values(): print(item)
        EnumItem(0, 'func', doc=Doc(title='Functional Mode'))
        EnumItem(1, 'mbist', doc=Doc(title='Memory BIST Mode'))
    """

    keytype: u.BitType = u.BitType()

    comment: str = "Memory Built-In Self-Test"

    def _build(self) -> None:
        self._add(0, "func", "Functional Mode")
        self._add(1, "mbist", "Memory BIST Mode")


class DftModeType(u.AStructType):
    """
    DFT Modes.

        >>> for item in DftModeType().values(): print(item)
        StructItem('test_mode', TestModeType(), doc=Doc(title='Test Mode'...))
        StructItem('scan_mode', ScanModeType(), doc=Doc(title='Scan Mode'...))
        StructItem('scan_shift', ScanShiftType(), doc=Doc(title='Scan Shift'...))
        StructItem('mbist_mode', MbistModeType(), doc=Doc(title='Memory BIST Mode'...))
    """

    comment: str = "Test Control"

    def _build(self) -> None:
        self._add("test_mode", TestModeType(), title="Test Mode")
        self._add("scan_mode", ScanModeType(), title="Scan Mode")
        self._add("scan_shift", ScanShiftType(), title="Scan Shift")
        self._add("mbist_mode", MbistModeType(), title="Memory BIST Mode")


class JtagType(u.AStructType):
    """
    JTAG Type.

        >>> for item in JtagType().values(): print(item)
        StructItem('tms', BitType())
        StructItem('trst', BitType())
        StructItem('tdi', BitType())
        StructItem('tck', BitType())
        StructItem('tdo', BitType(), orientation=BWD)
        StructItem('tdo_oe_n', BitType(), orientation=BWD)
    """

    def _build(self) -> None:
        self._add("tms", u.BitType(), u.FWD)
        self._add("trst", u.BitType(), u.FWD)
        self._add("tdi", u.BitType(), u.FWD)
        self._add("tck", u.BitType(), u.FWD)
        self._add("tdo", u.BitType(), u.BWD)
        self._add("tdo_oe_n", u.BitType(), u.BWD)


class ScanDataType(u.AStructType):
    """
    Scan Data.

        >>> for item in ScanDataType().values(): print(item)
        StructItem('si0', BitType())
        StructItem('si1', BitType())
        StructItem('si2', BitType())
        StructItem('si3', BitType())
        StructItem('si4', BitType())
        StructItem('si5', BitType())
        StructItem('si6', BitType())
        StructItem('si7', BitType())
        StructItem('so0', BitType(), orientation=BWD)
        StructItem('so1', BitType(), orientation=BWD)
        StructItem('so2', BitType(), orientation=BWD)
        StructItem('so3', BitType(), orientation=BWD)
        StructItem('so4', BitType(), orientation=BWD)
        StructItem('so5', BitType(), orientation=BWD)
        StructItem('so6', BitType(), orientation=BWD)
        StructItem('so7', BitType(), orientation=BWD)

    """

    comment: str = "Scan Chain Data"
    width: int = 8

    def _build(self) -> None:
        for idx in range(self.width):
            self._add(f"si{idx}", u.BitType(), u.FWD)
        for idx in range(self.width):
            self._add(f"so{idx}", u.BitType(), u.BWD)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Find Module With Specific Attribute Or Method.

    >>> import ucdp as u
    >>> class SubMod(u.AMod):
    ...     def _build(self):
    ...         pass
    >>> class AnotherSubMod(u.AMod):
    ...     def _build(self):
    ...         pass
    ...     @property
    ...     def feature(self):
    ...         return "Hello World"
    >>> class TopMod(u.AMod):
    ...     def _build(self):
    ...         SubMod(self, 'u_sub')
    ...         AnotherSubMod(self, 'u_anothersub')

    >>> from ucdp_glbl.finder import find
    >>> top = TopMod()
    >>> find([top], 'feature')
    (<ucdp_glbl.finder.AnotherSubMod(inst='top/u_anothersub', ...)>, 'Hello World')
"""

from collections.abc import Iterable
from typing import Any

import ucdp as u


def find(mods: Iterable[u.BaseMod], name: str) -> tuple[u.BaseMod, Any]:
    """Find module with attribute named `name`, starting from top to bottom."""
    undefined = object()
    attrs: list[Any] = []
    attrmods: list[u.BaseMod] = []
    insts = list(mods)
    while insts and not attrs:
        subinsts = []
        # Search for 'name' on all modules.
        for inst in insts:
            subinsts.extend(inst.insts)
            attr = getattr(inst, name, undefined)
            if attr is not undefined:
                attrs.append(attr)
                attrmods.append(inst)
        # None of the modules had a 'name' property. So continue on next level
        if attrs:
            break
        insts = subinsts

    if not attrs:
        raise ValueError(f"No module found which implements {name!r}")

    if len(attrs) == 1:
        return attrmods[0], attrs[0]

    lines = [f"Multiple modules implement {name!r}:"]
    lines += [f"  {mod!r}" for mod in attrmods]
    lines.append(f"Implement {name!r} on a parent module or choose a different top.")
    raise ValueError("\n".join(lines))



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Memory Types.
"""

from typing import Literal, TypeAlias

import ucdp as u

SliceWidths: TypeAlias = tuple[int | u.Expr, ...]
"""Word Slice Widths."""

Addressing: TypeAlias = Literal["byte", "data"]
"""Address Mode."""


class MemIoType(u.AStructType):
    """
    Memory IO-Type.

    Attributes:
        datawidth: Data Width in Bits
        addrwidth: Address Width in Bits
        writable: Read-Only or Read/Writable Memory
        slicewidths: Word Slices for Modification, Sum MUST be identical to datawidth
        err: Report Access Error
        addressing: Addressing Mode

    ROM Example:

        >>> from ucdp_glbl.mem import MemIoType
        >>> iotype = MemIoType(datawidth=32, addrwidth=8, writable=False)
        >>> iotype
        MemIoType(datawidth=32, addrwidth=8, writable=False)
        >>> for item in iotype.values(): print(item)
        StructItem('ena', EnaType(), doc=Doc(title='Memory Access Enable'))
        StructItem('addr', UintType(8), doc=Doc(title='Memory Address'))
        StructItem('rdata', UintType(32), orientation=BWD, doc=Doc(title='Memory Read Data'))

    RAM Example:

        >>> from ucdp_glbl.mem import MemIoType
        >>> iotype = MemIoType(datawidth=32, addrwidth=8, writable=True)
        >>> iotype
        MemIoType(datawidth=32, addrwidth=8, writable=True)
        >>> for item in iotype.values(): print(item)
        StructItem('ena', EnaType(), doc=Doc(title='Memory Access Enable'))
        StructItem('addr', UintType(8), doc=Doc(title='Memory Address'))
        StructItem('wena', EnaType(), doc=Doc(title='Memory Write Enable'))
        StructItem('wdata', UintType(32), doc=Doc(title='Memory Write Data'))
        StructItem('rdata', UintType(32), orientation=BWD, doc=Doc(title='Memory Read Data'))

    RAM Example with Byte Access and error:

        >>> from ucdp_glbl.mem import MemIoType
        >>> iotype = MemIoType.with_slicewidth(datawidth=32, addrwidth=8, writable=True, slicewidth=8, err=True)
        >>> iotype
        MemIoType(datawidth=32, addrwidth=8, writable=True, slicewidths=(8, 8, 8, 8), err=True)
        >>> for item in iotype.values(): print(item)
        StructItem('ena', EnaType(), doc=Doc(title='Memory Access Enable'))
        StructItem('addr', UintType(8), doc=Doc(title='Memory Address'))
        StructItem('wena', EnaType(), doc=Doc(title='Memory Write Enable'))
        StructItem('wdata', UintType(32), doc=Doc(title='Memory Write Data'))
        StructItem('rdata', UintType(32), orientation=BWD, doc=Doc(title='Memory Read Data'))
        StructItem('sel', UintType(4), doc=Doc(title='Slice Selects'))
        StructItem('err', BitType(), orientation=BWD, doc=Doc(title='Memory Access Failed.'))

    Non-Slicable:

        >>> MemIoType.with_slicewidth(datawidth=31, addrwidth=8, writable=True, slicewidth=8)
        Traceback (most recent call last):
        ...
        ValueError: Cannot split 31 bits into slices of 8 bits

    """

    datawidth: int | u.Expr
    addrwidth: int | u.Expr
    writable: bool
    slicewidths: SliceWidths | None = None
    err: bool = False
    addressing: Addressing = "byte"

    def _build(self) -> None:
        datatype = u.UintType(self.datawidth)
        addrtype = u.UintType(self.addrwidth)
        self._add("ena", u.EnaType(), title="Memory Access Enable")
        self._add("addr", addrtype, title="Memory Address")
        if self.writable:
            self._add("wena", u.EnaType(), title="Memory Write Enable")
            self._add("wdata", datatype, title="Memory Write Data")
        self._add("rdata", datatype, title="Memory Read Data", orientation=u.BWD)
        if self.slicewidths:
            self._add("sel", u.UintType(len(self.slicewidths)), title="Slice Selects")
        if self.err:
            self._add("err", u.BitType(), title="Memory Access Failed.", orientation=u.BWD)

    @staticmethod
    def with_slicewidth(
        datawidth: int | u.Expr,
        addrwidth: int | u.Expr,
        writable: bool,
        slicewidth: int | u.Expr | None = None,
        err: bool = False,
    ) -> "MemIoType":
        """Create `MemIoType` based on given `slicewidth` instead of individual `slicewidths`."""
        slicewidths = calc_slicewidths(datawidth, slicewidth)
        return MemIoType(datawidth=datawidth, addrwidth=addrwidth, writable=writable, slicewidths=slicewidths, err=err)


def calc_slicewidths(datawidth: int | u.Expr, slicewidth: int | u.Expr | None = None) -> SliceWidths | None:
    """
    Calculate slicewidths.

        >>> calc_slicewidths(32) is None
        True
        >>> calc_slicewidths(32, slicewidth=8)
        (8, 8, 8, 8)
        >>> calc_slicewidths(32, slicewidth=5)
        Traceback (most recent call last):
          ...
        ValueError: Cannot split 32 bits into slices of 5 bits
    """
    if slicewidth is None:
        return None
    if datawidth % slicewidth != 0:
        raise ValueError(f"Cannot split {datawidth} bits into slices of {slicewidth} bits")
    slices = datawidth // slicewidth
    return (slicewidth,) * slices



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Interrupt Types.
"""

import ucdp as u


class LevelIrqType(u.AEnumType):
    """
    Level IRQ.

        >>> import ucdp_glbl
        >>> irq = ucdp_glbl.irq.LevelIrqType()
        >>> irq
        LevelIrqType()
        >>> irq.width
        1
        >>> for item in irq.values(): print(item)
        EnumItem(0, 'idle')
        EnumItem(1, 'active')
    """

    keytype: u.BitType = u.BitType()

    comment: str = "Level IRQ"

    def _build(self) -> None:
        self._add(0, "idle")
        self._add(1, "active")


class ToggleIrqType(u.BitType):
    """
    Toggle IRQ.

        >>> import ucdp_glbl
        >>> irq = ucdp_glbl.irq.LevelIrqType()
        >>> irq
        LevelIrqType()
        >>> irq.width
        1
    """

    comment: str = "Toggle IRQ"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Streaming Types.
"""

from typing import Any

import ucdp as u


class ValidType(u.AEnumType):
    """
    Valid Type.

    A signal of that type indicates if the related data is valid,
    driven by the transmitter.

        >>> import ucdp_glbl
        >>> valid = ucdp_glbl.stream.ValidType()
        >>> valid
        ValidType()
        >>> valid.width
        1
        >>> for item in valid.values(): print(item)
        EnumItem(0, 'invalid')
        EnumItem(1, 'valid')
    """

    keytype: u.BitType = u.BitType()

    def _build(self) -> None:
        self._add(0, "invalid")
        self._add(1, "valid")


class AcceptType(u.AEnumType):
    """
    Accept Type.

    A signal of that type indicates if the related data is taken over,
    driven by the receiver.

        >>> import ucdp_glbl
        >>> accept = ucdp_glbl.stream.AcceptType()
        >>> accept
        AcceptType()
        >>> accept.width
        1
        >>> for item in accept.values(): print(item)
        EnumItem(0, 'busy')
        EnumItem(1, 'ready')
    """

    keytype: u.BitType = u.BitType()

    def _build(self) -> None:
        self._add(0, "busy")
        self._add(1, "ready")


class AStreamType(u.AStructType):
    """
    Abstract Stream Type.

    A stream contains at least a `valid` and `accept` member.
    It is up to the child classes to define the corresponding data.
    """

    def _build(self) -> None:
        self._add("valid", ValidType(), u.FWD)
        self._add("accept", AcceptType(), u.BWD)


class StreamType(AStreamType):
    """
    Simple Stream Type with one `data` signal.

        >>> import ucdp_glbl
        >>> stream = ucdp_glbl.stream.StreamType(8)
        >>> for item in stream.values(): print(item)
        StructItem('valid', ValidType())
        StructItem('accept', AcceptType(), orientation=BWD)
        StructItem('data', UintType(8))
    """

    width: int | u.Expr

    def __init__(self, width: int | u.Expr, **kwargs: dict[str, Any]) -> None:
        super().__init__(width=width, **kwargs)

    def _build(self) -> None:
        super()._build()
        self._add("data", u.UintType(self.width))


class AnyStreamType(AStreamType):
    """
    Any Stream Type.

        >>> import ucdp_glbl
        >>> stream = ucdp_glbl.stream.AnyStreamType(u.SintType(16))
        >>> for item in stream.values(): print(item)
        StructItem('valid', ValidType())
        StructItem('accept', AcceptType(), orientation=BWD)
        StructItem('data', SintType(16))
    """

    type_: u.BaseType

    def __init__(self, type_: u.BaseType, **kwargs: dict[str, Any]) -> None:
        super().__init__(type_=type_, **kwargs)

    def _build(self) -> None:
        super()._build()
        self._add("data", self.type_)
